@@ -9,7 +9,6 @@ use crate::path::{Path, PathBuf};
 use crate::sys::time::SystemTime;
 use crate::sys::{helpers, unsupported};
 
-#[expect(dead_code)]
 const FILE_PERMISSIONS_MASK: u64 = r_efi::protocols::file::READ_ONLY;
 
 pub struct File(!);
@@ -18,9 +17,8 @@ pub struct File(!);
 pub struct FileAttr {
     attr: u64,
     size: u64,
-    accessed: SystemTime,
-    modified: SystemTime,
-    created: SystemTime,
+    file_time: FileTimes,
+    created: Option<SystemTime>,
 }
 
 pub struct ReadDir(!);
@@ -66,15 +64,20 @@ impl FileAttr {
     }
 
     pub fn modified(&self) -> io::Result<SystemTime> {
-        Ok(self.modified)
+        self.file_time
+            .modified
+            .ok_or(io::const_error!(io::ErrorKind::InvalidData, "modification time is not valid"))
     }
 
     pub fn accessed(&self) -> io::Result<SystemTime> {
-        Ok(self.accessed)
+        self.file_time
+            .accessed
+            .ok_or(io::const_error!(io::ErrorKind::InvalidData, "last access time is not valid"))
     }
 
     pub fn created(&self) -> io::Result<SystemTime> {
-        Ok(self.created)
+        self.created
+            .ok_or(io::const_error!(io::ErrorKind::InvalidData, "creation time is not valid"))
     }
 
     fn from_uefi(info: helpers::UefiBox<file::Info>) -> Self {
@@ -82,8 +85,10 @@ impl FileAttr {
             Self {
                 attr: (*info.as_ptr()).attribute,
                 size: (*info.as_ptr()).file_size,
-                modified: uefi_fs::uefi_to_systemtime((*info.as_ptr()).modification_time),
-                accessed: uefi_fs::uefi_to_systemtime((*info.as_ptr()).last_access_time),
+                file_time: FileTimes {
+                    modified: uefi_fs::uefi_to_systemtime((*info.as_ptr()).modification_time),
+                    accessed: uefi_fs::uefi_to_systemtime((*info.as_ptr()).last_access_time),
+                },
                 created: uefi_fs::uefi_to_systemtime((*info.as_ptr()).create_time),
             }
         }
@@ -103,7 +108,6 @@ impl FilePermissions {
         Self(attr & r_efi::protocols::file::READ_ONLY != 0)
     }
 
-    #[expect(dead_code)]
     const fn to_attr(&self) -> u64 {
         if self.0 { r_efi::protocols::file::READ_ONLY } else { 0 }
     }
@@ -360,16 +364,40 @@ pub fn rename(_old: &Path, _new: &Path) -> io::Result<()> {
     unsupported()
 }
 
-pub fn set_perm(_p: &Path, _perm: FilePermissions) -> io::Result<()> {
-    unsupported()
+pub fn set_perm(p: &Path, perm: FilePermissions) -> io::Result<()> {
+    let f = uefi_fs::File::from_path(p, file::MODE_READ | file::MODE_WRITE, 0)?;
+    let mut file_info = f.file_info()?;
+
+    unsafe {
+        (*file_info.as_mut_ptr()).attribute =
+            ((*file_info.as_ptr()).attribute & !FILE_PERMISSIONS_MASK) | perm.to_attr()
+    };
+
+    f.set_file_info(file_info)
 }
 
-pub fn set_times(_p: &Path, _times: FileTimes) -> io::Result<()> {
-    unsupported()
+pub fn set_times(p: &Path, times: FileTimes) -> io::Result<()> {
+    // UEFI does not support symlinks
+    set_times_nofollow(p, times)
 }
 
-pub fn set_times_nofollow(_p: &Path, _times: FileTimes) -> io::Result<()> {
-    unsupported()
+pub fn set_times_nofollow(p: &Path, times: FileTimes) -> io::Result<()> {
+    let f = uefi_fs::File::from_path(p, file::MODE_READ | file::MODE_WRITE, 0)?;
+    let mut file_info = f.file_info()?;
+
+    if let Some(x) = times.accessed {
+        unsafe {
+            (*file_info.as_mut_ptr()).last_access_time = uefi_fs::systemtime_to_uefi(x);
+        }
+    }
+
+    if let Some(x) = times.modified {
+        unsafe {
+            (*file_info.as_mut_ptr()).modification_time = uefi_fs::systemtime_to_uefi(x);
+        }
+    }
+
+    f.set_file_info(file_info)
 }
 
 pub fn rmdir(p: &Path) -> io::Result<()> {
@@ -554,6 +582,17 @@ mod uefi_fs {
             if r.is_error() { Err(io::Error::from_raw_os_error(r.as_usize())) } else { Ok(info) }
         }
 
+        pub(crate) fn set_file_info(&self, mut info: UefiBox<file::Info>) -> io::Result<()> {
+            let file_ptr = self.0.as_ptr();
+            let mut info_id = file::INFO_ID;
+
+            let r = unsafe {
+                ((*file_ptr).set_info)(file_ptr, &mut info_id, info.len(), info.as_mut_ptr().cast())
+            };
+
+            if r.is_error() { Err(io::Error::from_raw_os_error(r.as_usize())) } else { Ok(()) }
+        }
+
         pub(crate) fn delete(self) -> io::Result<()> {
             let file_ptr = self.0.as_ptr();
             let r = unsafe { ((*file_ptr).delete)(file_ptr) };
@@ -627,9 +666,9 @@ mod uefi_fs {
 
     /// EDK2 FAT driver uses EFI_UNSPECIFIED_TIMEZONE to represent localtime. So for proper
     /// conversion to SystemTime, we use the current time to get the timezone in such cases.
-    pub(crate) fn uefi_to_systemtime(mut time: r_efi::efi::Time) -> SystemTime {
+    pub(crate) fn uefi_to_systemtime(mut time: r_efi::efi::Time) -> Option<SystemTime> {
         time.timezone = if time.timezone == r_efi::efi::UNSPECIFIED_TIMEZONE {
-            time::system_time_internal::now().unwrap().timezone
+            time::system_time_internal::now().timezone
         } else {
             time.timezone
         };
@@ -637,9 +676,8 @@ mod uefi_fs {
     }
 
     /// Convert to UEFI Time with the current timezone.
-    #[expect(dead_code)]
-    fn systemtime_to_uefi(time: SystemTime) -> r_efi::efi::Time {
-        let now = time::system_time_internal::now().unwrap();
+    pub(crate) fn systemtime_to_uefi(time: SystemTime) -> r_efi::efi::Time {
+        let now = time::system_time_internal::now();
         time.to_uefi_loose(now.timezone, now.daylight)
     }
 }
